@@ -20,9 +20,10 @@ except ImportError:
     from ._child import wait_for_child
 
 import logging
+
 logger = logging.getLogger(__name__)
 
-_mswindows = (sys.platform == "win32")
+_mswindows = sys.platform == "win32"
 
 try:
     _wait_readable = trio.lowlevel.wait_readable
@@ -107,8 +108,10 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
     Arguments:
         queue_len:
             The maximum length of the internal event queue.
-            The default is 1000. Use more for large programs,
-            or when running benchmarks.
+            The default of None means unlimited. A limit should be specified
+            only if you would rather crash your program than use too
+            much memory, because it's not feasible to enforce graceful
+            backpressure here.
     """
 
     # This code implements a semi-efficient way to run asyncio code within Trio.
@@ -140,7 +143,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
 
     def __init__(self, queue_len=None):
         if queue_len is None:
-            queue_len = 10000
+            queue_len = math.inf
 
         # Processing queue
         self._q_send, self._q_recv = trio.open_memory_channel(queue_len)
@@ -170,27 +173,27 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
     def __repr__(self):
         try:
             return "<%s running=%s at 0x%x>" % (
-                self.__class__.__name__, "closed" if self._closed else "no"
-                if self._stopped.is_set() else "yes", id(self)
+                self.__class__.__name__,
+                "closed" if self._closed else "no" if self._stopped.is_set() else "yes",
+                id(self),
             )
         except Exception as exc:
             return "<%s ?:%s>" % (self.__class__.__name__, repr(exc))
 
     def set_default_executor(self, executor):
         if not isinstance(executor, TrioExecutor):
-            if 'pytest' not in sys.modules:
+            if "pytest" not in sys.modules:
                 warnings.warn(
                     "executor %r is not supported by trio-asyncio and will "
                     "not be used" % (repr(executor),),
                     RuntimeWarning,
-                    stacklevel=2
+                    stacklevel=2,
                 )
             return
         super().set_default_executor(executor)
 
     def time(self):
-        """Return Trio's idea of the current time.
-        """
+        """Return Trio's idea of the current time."""
         if self._task is None:
             raise RuntimeError("This loop is closed.")
         return self._task._runner.clock.current_time()
@@ -297,7 +300,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         :param callback: Sync function to call.
         :return: a handle which may be used to cancel the timer.
         """
-        self._check_callback(callback, 'call_later')
+        self._check_callback(callback, "call_later")
         assert delay >= 0, delay
         h = asyncio.TimerHandle(delay + self.time(), callback, args, self, **context)
         self._queue_handle(h)
@@ -308,15 +311,17 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
 
         Note that the callback is a sync function.
         """
-        self._check_callback(callback, 'call_at')
-        return self._queue_handle(asyncio.TimerHandle(when, callback, args, self, **context))
+        self._check_callback(callback, "call_at")
+        return self._queue_handle(
+            asyncio.TimerHandle(when, callback, args, self, **context)
+        )
 
     def call_soon(self, callback, *args, **context):
         """asyncio's defer-to-mainloop callback executor.
 
         Note that the callback is a sync function.
         """
-        self._check_callback(callback, 'call_soon')
+        self._check_callback(callback, "call_soon")
         return self._queue_handle(asyncio.Handle(callback, args, self, **context))
 
     def call_soon_threadsafe(self, callback, *args, **context):
@@ -324,7 +329,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
 
         Note that the callback is a sync function.
         """
-        self._check_callback(callback, 'call_soon_threadsafe')
+        self._check_callback(callback, "call_soon_threadsafe")
         self._check_closed()
         h = asyncio.Handle(callback, args, self, **context)
         self._token.run_sync_soon(self._q_send.send_nowait, h)
@@ -355,7 +360,16 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
     if not _mswindows:
 
         async def _make_subprocess_transport(
-                self, protocol, args, shell, stdin, stdout, stderr, bufsize, extra=None, **kwargs
+            self,
+            protocol,
+            args,
+            shell,
+            stdin,
+            stdout,
+            stderr,
+            bufsize,
+            extra=None,
+            **kwargs
         ):
             """Make a subprocess transport. Asyncio context."""
 
@@ -407,7 +421,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
 
         Returns an asyncio.Future.
         """
-        self._check_callback(func, 'run_in_executor')
+        self._check_callback(func, "run_in_executor")
         self._check_closed()
         if executor is None:  # pragma: no branch
             executor = self._default_executor
@@ -437,21 +451,20 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         self._token.run_sync_soon(self._q_send.send_nowait, h)
 
     def add_signal_handler(self, sig, callback, *args):
-        """asyncio's method to add a signal handler.
-        """
+        """asyncio's method to add a signal handler."""
         self._check_closed()
         self._check_signal(sig)
         if sig == signal.SIGKILL:
             raise RuntimeError("SIGKILL cannot be caught")
         h = asyncio.Handle(callback, args, self)
-        assert sig not in self._signal_handlers, \
-            "Signal %d is already being caught" % (sig,)
+        assert sig not in self._signal_handlers, "Signal %d is already being caught" % (
+            sig,
+        )
         self._orig_signals[sig] = signal.signal(sig, self._handle_sig)
         self._signal_handlers[sig] = h
 
     def remove_signal_handler(self, sig):
-        """asyncio's method to remove a signal handler.
-        """
+        """asyncio's method to remove a signal handler."""
         # self._check_signal(sig)
         try:
             h = self._signal_handlers.pop(sig)
@@ -481,15 +494,45 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         self._ensure_fd_no_transport(fd)
         return self._add_reader(fd, callback, *args)
 
-    def _add_reader(self, fd, callback, *args):
+    # Local helper to factor out common logic between _add_reader/_add_writer
+    def _add_io_handler(self, set_handle, wait_ready, fd, callback, args):
         self._check_closed()
         handle = ScopedHandle(callback, args, self)
-        reader = self._set_read_handle(fd, handle)
-        if reader is not None:
-            reader.cancel()
+        old_handle = set_handle(fd, handle)
+
+        if old_handle is not None:
+            old_handle.cancel()
         if self._token is None:
-            return
-        self._nursery.start_soon(self._reader_loop, fd, handle)
+            return None
+        self._nursery.start_soon(self._io_task, fd, handle, wait_ready)
+        return handle
+
+    async def _io_task(self, fd, handle, wait_ready):
+        with handle._scope:
+            try:
+                while True:
+                    if handle._cancelled:
+                        break
+                    try:
+                        await wait_ready(fd)
+                    except OSError:
+                        # maybe someone did
+                        #   h = add_reader(sock); h.cancel(); sock.close()
+                        # without yielding to the event loop
+                        if handle._cancelled:
+                            break
+                        raise
+                    if handle._cancelled:
+                        break
+                    handle._run()
+                    await self.synchronize()
+            except Exception as exc:
+                handle._raise(exc)
+
+    def _add_reader(self, fd, callback, *args):
+        return self._add_io_handler(
+            self._set_read_handle, _wait_readable, fd, callback, args
+        )
 
     def _set_read_handle(self, fd, handle):
         try:
@@ -502,20 +545,6 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
             self._selector.modify(fd, mask | EVENT_READ, (handle, writer))
             return reader
 
-    async def _reader_loop(self, fd, handle):
-        with handle._scope:
-            try:
-                while True:
-                    if handle._cancelled:
-                        break
-                    await _wait_readable(fd)
-                    if handle._cancelled:
-                        break
-                    handle._run()
-                    await self.synchronize()
-            except Exception as exc:
-                handle._raise(exc)
-
     # writing to a file descriptor
 
     def add_writer(self, fd, callback, *args):
@@ -535,15 +564,10 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
 
     # remove_writer: unchanged from asyncio
 
-    def _add_writer(self, fd, callback, *args):
-        self._check_closed()
-        handle = ScopedHandle(callback, args, self)
-        writer = self._set_write_handle(fd, handle)
-        if writer is not None:
-            writer.cancel()
-        if self._token is None:
-            return
-        self._nursery.start_soon(self._writer_loop, fd, handle)
+    def _add_writer(self, fd, callback, *args, _defer_start=False):
+        return self._add_io_handler(
+            self._set_write_handle, _wait_writable, fd, callback, args
+        )
 
     def _set_write_handle(self, fd, handle):
         try:
@@ -555,20 +579,6 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
             self._selector.modify(fd, mask | EVENT_WRITE, (reader, handle))
             return writer
 
-    async def _writer_loop(self, fd, handle):
-        with handle._scope:
-            try:
-                while True:
-                    if handle._cancelled:
-                        break
-                    await _wait_writable(fd)
-                    if handle._cancelled:
-                        break
-                    handle._run()
-                    await self.synchronize()
-            except Exception as exc:
-                handle._raise(exc)
-
     def autoclose(self, fd):
         """
         Mark a file descriptor so that it's auto-closed along with this loop.
@@ -581,7 +591,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         :param fd: Either an integer (Unix file descriptor) or an object
                    with a ``fileno`` method providing one.
         """
-        if hasattr(fd, 'fileno'):
+        if hasattr(fd, "fileno"):
             fd = fd.fileno()
         self._close_files.add(fd)
 
@@ -597,7 +607,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
                    with a ``fileno()`` method providing one.
         :raises KeyError: if the descriptor is not marked to be auto-closed.
         """
-        if hasattr(fd, 'fileno'):
+        if hasattr(fd, "fileno"):
             fd = fd.fileno()
         self._close_files.remove(fd)
 
@@ -741,6 +751,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         # clean core fields
         self._nursery = None
         self._task = None
+        self._token = None
 
     def is_running(self):
         if self._stopped is None:
@@ -748,8 +759,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         return not self._stopped.is_set()
 
     def run_forever(self):
-        """asyncio's method to run the loop until it is stopped.
-        """
+        """asyncio's method to run the loop until it is stopped."""
         raise RuntimeError("This is not a sync loop")
 
     def run_until_complete(self, coro_or_future):
@@ -768,6 +778,11 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         """
         await self._stopped.wait()
 
+    def _trio_io_cancel(self, cancel_scope):
+        """Called when a ScopedHandle representing an I/O reader or writer
+        has its cancel() method called."""
+        cancel_scope.cancel()
+
     def stop(self):
         """Halt the main loop.
 
